/*
** Module   : process.c
** Abstract : implementation for process creation/waiting logic in libp2j.so 
**            native library which is called via JNI
**
** Copyright (c) 2005-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------Description----------------------
** 001 LMR 20101123 Initial revision of this file. The code was moved from 
**                  Toolkit.c in the charva package (built as libTerminal.so).
**                  It was an original work created by Golden Code and was 
**                  never a part of the Rob Pitman charva distribution.  For
**                  this reason, it is fully owned by Golden Code and has been
**                  moved here to keep it separate from the LGPL charva code.
** 002 GES 20110629 Removed unused var reference that was causing a compile
**                  warning.
** 003 EVL 20130329 Making changes to compile Windows DLL.
*/ 

#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <errno.h>
#ifndef WIN_TARGET
#include <sys/wait.h>
#include <term.h>
#endif

#include "com_goldencode_p2j_util_LaunchManager.h"

// cache field IDs for performance
static jfieldID jfd    = NULL;

/**
 * initializeFileDescriptor - initialize the passed file descriptor object 
 *                            with the pty fd
 *
 * @param   env         pointer to an object providing access to JNI environment

 * @param   descriptor  object representing the FileDescriptor Java instance 
 *
 * @param   fd          the pty file descriptor
 *
 */ 
void initializeFileDescriptor(JNIEnv* env, jobject descriptor, int fd)
{
   if (jfd == NULL)
   {
      jclass fdcls = (*env)->GetObjectClass(env, descriptor);

      jfd = (*env)->GetFieldID(env, fdcls, "fd", "I");
   }

   (*env)->SetIntField(env, descriptor, jfd, fd);
}

/**
 * init_Linux - Linux version of set up signal handling to ignore job control signals
 *              (reflecteds in the child process as well)
 *
 * @param   env   pointer to an object providing access to JNI environment
 *
 * @param   cls   reference to the Java class who called this function
 *                (from static context)
 *
 */ 
#ifndef WIN_TARGET
JNIEXPORT void JNICALL init_Linux(JNIEnv* env, jclass class)
{
   // ignore SIGINT, SIGTSTP, SIGTTIN, SIGTTOU signals
   // this is done once and early enough and these signals remain ignored
   struct sigaction sa;
   sa.sa_handler = SIG_IGN;
   sa.sa_flags = 0;
   sigemptyset(&sa.sa_mask);

   sigaction(SIGINT,  &sa, NULL); 
   sigaction(SIGTSTP, &sa, NULL);
   sigaction(SIGTTIN, &sa, NULL);
   sigaction(SIGTTOU, &sa, NULL);
}
#endif

/**
 * init_Win - Windows version of set up signal handling to ignore job control signals
 *            (reflecteds in the child process as well)
 *
 * @param   env   pointer to an object providing access to JNI environment
 *
 * @param   cls   reference to the Java class who called this function
 *                (from static context)
 *
 */ 
#ifdef WIN_TARGET
JNIEXPORT void JNICALL init_Win(JNIEnv* env, jclass class)
{
}
#endif

/**
 * Java_com_goldencode_p2j_util_LaunchManager_init - 
 *                         set up signal handling to ignore job control
 *                         signals (reflecteds in the child process as well)
 *
 * @param   env   pointer to an object providing access to JNI environment
 *
 * @param   cls   reference to the Java class who called this function
 *                (from static context)
 *
 */ 
JNIEXPORT void JNICALL Java_com_goldencode_p2j_util_LaunchManager_init(
                                                                JNIEnv* env,
                                                                jclass class)
{
#ifdef WIN_TARGET
   init_Win(env, class);
#else
   init_Linux(env, class);
#endif
}

/**
 * pseudoTerminalLaunch_Linux - Linux version of spawning a child process redirecting its standard
 *                              input, output and error to particular file descriptors
 *                              (FileDescriptor Java instances) 
 *      
 *
 * @param   env       pointer to a structure that provides access to JNI
 *                    environment
 *
 * @param   cls       reference to the Java class who called this function
 *                    (from static context)
 *
 * @param   cmdline   the command line used for launching the child process  
 *
 * @param   in        FileDescriptor Java object for stdin
 *
 * @param   out       FileDescriptor Java object for stdout
 *
 * @param   err       FileDescriptor Java object for stderr
 *
 *
 * @param   silent  boolean flag controlling interactive mode
 *
 * @return  > 1  PID of spawned child process if successful
 *          < 0  on error
 */ 

#ifndef WIN_TARGET
JNIEXPORT jint JNICALL
   pseudoTerminalLaunch_Linux(JNIEnv* env, jclass cls, jobjectArray cmdline, jobject in,
                              jobject out, jobject err, jboolean silent)
{
   jint pid;
   
   // make sure we have valid argument references
   if (cmdline == NULL)
   {
      return -1;
   }
   
   // make sure we have a tty as stdin
   if (!isatty(STDIN_FILENO))
   {
      return -2;
   }
   
   // access redirection flags and create pipes
   int pipe_i[2];                // pipe for stdin
   if (in != NULL)
   {
      if (pipe(pipe_i) < 0)
      {
         return -3;
      }
   }
   
   int pipe_o[2];                // pipe for stdout
   if (out != NULL)
   {
      if (pipe(pipe_o) < 0)
      {
         if (in != NULL)
         {
            close(pipe_i[0]);
            close(pipe_i[1]);
         }
         return -4;
      }
   }
   
   int pipe_e[2];                // pipe for stderr
   if (err != NULL)
   {
      if (pipe(pipe_e) < 0)
      {
         if (in != NULL)
         {
            close(pipe_i[0]);
            close(pipe_i[1]);
         }
         if (out != NULL)
         {
            close(pipe_o[0]);
            close(pipe_o[1]);
         }
         return -5;
      }
   }

   // due to a bug in the JVM 1.4.2 and earlier, we must override any
   // signal handler for SIGCHLD that may exist because otherwise the JVM
   // may reap our process instead of letting us do it (see Sun J2SE bugs 
   // #4945203, #4785154 and #4763362)
   signal(SIGCHLD, SIG_IGN);

   // create a child process
   pid = (jint) fork();
   
   if (pid < 0)
   {
      // error, nothing more to do
      return -6;
   }
   else if (pid == 0)
   {
      // child process

      if (!silent)
      {
         // we have to leave the screen clear for the program being launched
         system("tput clear");

         // in interactive mode, we redirect stderr to stdout
         dup2(STDOUT_FILENO, STDERR_FILENO);
      }

      // initialize the file descriptors for the redirected handles
      if (in != NULL)
      {
         // don't need write end for child's stdin in child
         close(pipe_i[1]);
         
         // redirect this pipe to stdin
         if (pipe_i[0] != STDIN_FILENO)
         {
            if (dup2(pipe_i[0], STDIN_FILENO) != STDIN_FILENO)
            {
               // handle duplication problem
               exit(-1);
            }
            
            close(pipe_i[0]);    // not required after duplication
         }
      }
      
      if (out != NULL)
      {
         // don't need read end for child's stdout in child
         close(pipe_o[0]);
         
         // redirect this pipe to stdout
         if (pipe_o[1] != STDOUT_FILENO)
         {
            if (dup2(pipe_o[1], STDOUT_FILENO) != STDOUT_FILENO)
            {
               // handle duplication problem
               exit(-1);
            }
            
            close(pipe_o[1]);    // not required after duplication
         }
      }
      
      if (err != NULL)
      {
         // don't need read end for child's stderr in child
         close(pipe_e[0]);
         
         // redirect this pipe to stderr
         if (pipe_e[1] != STDERR_FILENO)
         {
            if (dup2(pipe_e[1], STDERR_FILENO) != STDERR_FILENO)
            {
               // handle duplication problem
               exit(-1);
            }
            
            close(pipe_e[1]);    // not required after duplication
         }
      }

      // must build the parameter list
      int len = (int) (*env)->GetArrayLength(env, cmdline);
      int i;
      
      char** cmds = malloc(sizeof(char*) * (len + 1)); 
      char** current = cmds;
      
      // copy the strings into our own array
      for (i = 0; i < len; i++, current++)
      {
         jobject next = (*env)->GetObjectArrayElement(env, cmdline, i); 
         char*   tmp  = (char*) (*env)->GetStringUTFChars(env,
                                                          (jstring) next,
                                                          NULL);
         
         if (tmp == NULL)
         {
            // out of memory!
            exit(-1);
         }
         
         *current = strdup(tmp);
         
         (*env)->ReleaseStringUTFChars(env, (jstring) next, tmp);
      }
      
      // terminate the argv[] with a null pointer (current will be pointing
      // at cmds[len] after the loop ends)
      *current = NULL;

      // exec (never returns unless there is an error)
      int rc = execvp(*cmds, cmds);
      
      // if we are here, there was a problem!
      fprintf(stderr, "ERROR: execvp rc = %i; errno = %i\n", rc, errno);
      exit(-2);
   }
   else
   {
      // parent process
      // initialize the file descriptors for the redirected handles
      if (in != NULL)
      {
         // don't need read end for child's stdin in parent
         close(pipe_i[0]);
         
         // pass the write end back
         initializeFileDescriptor(env, in, pipe_i[1]); 
      }
      
      if (out != NULL)
      {
         // don't need write end for child's stdout in parent
         close(pipe_o[1]);
         
         // pass the read end back
         initializeFileDescriptor(env, out, pipe_o[0]); 
      }
      
      if (err != NULL)
      {
         // don't need write end for child's stderr in parent
         close(pipe_e[1]);
         
         // pass the read end back
         initializeFileDescriptor(env, err, pipe_e[0]); 
      }
   }
   
   return pid;
}
#endif

/**
 * pseudoTerminalLaunch_Win - Windows version of spawning a child process redirecting its standard
 *                            input, output and error to particular file descriptors
 *                            (FileDescriptor Java instances) 
 *      
 *
 * @param   env       pointer to a structure that provides access to JNI
 *                    environment
 *
 * @param   cls       reference to the Java class who called this function
 *                    (from static context)
 *
 * @param   cmdline   the command line used for launching the child process  
 *
 * @param   in        FileDescriptor Java object for stdin
 *
 * @param   out       FileDescriptor Java object for stdout
 *
 * @param   err       FileDescriptor Java object for stderr
 *
 *
 * @param   silent  boolean flag controlling interactive mode
 *
 * @return  > 1  PID of spawned child process if successful
 *          < 0  on error
 */ 

#ifdef WIN_TARGET
JNIEXPORT jint JNICALL
   pseudoTerminalLaunch_Win(JNIEnv* env, jclass cls, jobjectArray cmdline, jobject in,
                            jobject out, jobject err, jboolean silent)
{
   jint pid = 0;

   return pid;
}
#endif

/**
 * Java_com_goldencode_p2j_util_LaunchManager_pseudoTerminalLaunch - 
 *      spawns a child process redirecting its standard input, output and
 *      error to particular file descriptors (FileDescriptor Java instances) 
 *      
 *
 * @param   env       pointer to a structure that provides access to JNI
 *                    environment
 *
 * @param   cls       reference to the Java class who called this function
 *                    (from static context)
 *
 * @param   cmdline   the command line used for launching the child process  
 *
 * @param   in        FileDescriptor Java object for stdin
 *
 * @param   out       FileDescriptor Java object for stdout
 *
 * @param   err       FileDescriptor Java object for stderr
 *
 *
 * @param   silent  boolean flag controlling interactive mode
 *
 * @return  > 1  PID of spawned child process if successful
 *          < 0  on error
 */ 

JNIEXPORT jint JNICALL
   Java_com_goldencode_p2j_util_LaunchManager_pseudoTerminalLaunch(JNIEnv* env,
                                                jclass        cls,
                                                jobjectArray  cmdline,
                                                jobject       in,
                                                jobject       out,
                                                jobject       err,
                                                jboolean      silent)
   
{
#ifdef WIN_TARGET
   return pseudoTerminalLaunch_Win(env, cls, cmdline, in, out, err, silent);
#else
   return pseudoTerminalLaunch_Linux(env, cls, cmdline, in, out, err, silent);
#endif
}

/**
 * pseudoTerminalWait_Linux - The Linux version of waits for a child process to terminate. 
 *
 * @param   env     pointer to a structure that provides access to JNI environment
 *
 * @param   cls     reference to the Java class who called this function from
 *                  a static context
 *
 * @param   pid     PID of the process to be waited
 *
 * @param   silent  boolean flag to control interactive mode
 *
 */ 
#ifndef WIN_TARGET
JNIEXPORT void JNICALL pseudoTerminalWait_Linux(JNIEnv* env, jclass cls, int pid, jboolean silent)
{
   pid_t child = (pid_t) pid;
   int   status;

   waitpid(child, &status, 0);

   if (!silent)
   {
      // in interactive mode we're calling functions in NCURSES directly, 
      // as opposed to executing command line utilities for the same purpose
      // (NCURSES library is a requirement anyway so we want to have a more
      // efficient and more controlled use of that interface as well, while 
      // also avoiding the hard requirement that other utility programs be
      // installed in addition to P2J)

      // put "Press space bar to continue." message on the last line

      // move cursor to the beginning of the last terminal line 
      char *s;
      s = tparm(tigetstr("cup"), tigetnum("lines"), 0, 0, 0, 0, 0, 0, 0, 0);
      putp(s);

      // display prompt
      printf("Press space bar to continue.");

      // move cursor to the end of the last terminal line (as in Progress)
      s = tparm(tigetstr("cup"), tigetnum("lines"), tigetnum("cols"),
                0, 0, 0, 0, 0, 0, 0);
      putp(s);
   }
}
#endif

/**
 * pseudoTerminalWait_Win - The Windows version of waits for a child process to terminate. 
 *
 * @param   env     pointer to a structure that provides access to JNI environment
 *
 * @param   cls     reference to the Java class who called this function from
 *                  a static context
 *
 * @param   pid     PID of the process to be waited
 *
 * @param   silent  boolean flag to control interactive mode
 *
 */ 
#ifdef WIN_TARGET
JNIEXPORT void JNICALL pseudoTerminalWait_Win(JNIEnv* env, jclass cls, jint pid, jboolean silent)
{
}
#endif

/**
 * Java_com_goldencode_p2j_util_LaunchManager_pseudoTerminalWait - 
 *          waits for a child process to terminate. 
 *
 * @param   env     pointer to a structure that provides access to JNI environment
 *
 * @param   cls     reference to the Java class who called this function from
 *                  a static context
 *
 * @param   pid     PID of the process to be waited
 *
 * @param   silent  boolean flag to control interactive mode
 *
 */ 
JNIEXPORT void JNICALL
   Java_com_goldencode_p2j_util_LaunchManager_pseudoTerminalWait(JNIEnv*  env,
                                                                 jclass   cls,
                                                                 jint      pid,
                                                             jboolean silent)
{
#ifdef WIN_TARGET
   pseudoTerminalWait_Win(env, cls, pid, silent);
#else
   pseudoTerminalWait_Linux(env, cls, pid, silent);
#endif
}

